/****************************************************************************
 Implementierung eines Algorithmus zum Parsen von TAGs mit Constraints

 Original: Vijay-Shankar (Dissertation)

 Implementierung (c) 1996 by Sven Thiel

 Benutzte Bibliotheken (auer Standard): LEDA-R-3.4

 Datei: vs.cc     
 ****************************************************************************

 Anforderungen an die Grammatik:
 - Grammatik ist in 2-Form
 - keine Epsilon-Bltter erlaubt
 - jeder Baum hat mind. ein Terminalblatt (vgl. gram_term)

 ****************************************************************************

 nderungen gegenber dem Original-Algorithmus:
 1 : Bedingungen fr die Indices in case2 gendert (nun symm. zu case1)
      (Gegenbsp.: gram_c2)
 2 : Schleifen in step 3-5 laufen jetzt bis n
      (Gegenbsp.: gram_n)
 3 : Bedingungen fr die Indices in case5 gendert und zustzl. Aufruf von 
      case4 (Gegenbsp.: gram_c5)
 4 : case3 wird nur noch fr j==k ausgefhrt 
      (spart etwas Zeit und vermeidet falsche Ausgaben)
 5 : case1 und case2 werden nur noch fr j != k ausgefhrt
      (spart Zeit und vermeidet unzulssige Aufnahmen von Knoten ins Array
       vgl. gram_n -> A[0,1,1,1])
 6 : Innere Schleife in step 3-5 luft nun von i+1 statt von i
      (spart Zeit und vermeidet unntige Aufnahmen von Knoten ins Array)

 ****************************************************************************/

#define VERSION "4.31"

#include <iostream.h>
#include <string.h>
#include <fstream.h>
#include <LEDA/set.h>

// *** Reprsentation der Grammatik: ***

char gram_name[50];

enum { noth, term, foot, root_aux, root_ini };
struct node {
  char symbol;       // Label des Knotens
  int left_sibl;     // Nr. des linken Bruders (0, falls nicht ex.)  
  int right_sibl;    // Nr. des rechten  "              "
  int father;        // Nr. des Vaters                  "
  int type;          // 0 = nichts besonderes, 1 = Terminalknoten, 
                     // 2 = Fu, 3 = Wurzel eines Auxiliarbaumes
                     // 4 = Wurzel eines Initialbaumes
  int anc_of_foot;   // 0 = kein Vorgnger des Fuknotens (bzw. Initialbaum)
  int OA;            // OA:  0 = keines, 1 = OA-Constraint 
  set<int>* SA;      // Menge der adjungierbaren Wurzelknoten von aux. Bumen
                     // (Falls alle Bume adjungierbar sind ist der Wert nil)

  ~node() { delete SA; }
};

// liest einen Knoten in der Grammatik aus einem Stream
istream& operator>> (istream& in, node& n)
{
  int SA_Flag;

  in >> n.symbol;          //cerr << n.symbol;
  in >> n.left_sibl;       //cerr << n.left_sibl;
  in >> n.right_sibl;      //cerr << n.right_sibl;
  in >> n.father;          //cerr << n.father;
  in >> n.type;            //cerr << n.type;
  in >> n.anc_of_foot;     //cerr << n.anc_of_foot;
  in >> n.OA;              //cerr << n.OA;
  in >> SA_Flag;           //cerr << SA_Flag << ':';
  
  if ( SA_Flag >= 0 ) {
    n.SA = new set<int>;
    int j, cn;
    for(j=1; j<=SA_Flag; j++) {
      in >> cn;            //cerr << cn;
      n.SA->insert(cn);
    }
  } 
  else 
    n.SA = nil;  
                           //cerr << '\n';  
}


int no_num;
node* Nodes; 

// *** Reprsentation der Eintrge in der A[i,j,k,l]-Matrix ***

// Die Klasse no_cs reprsentiert das Tupel (node,OA,SA), d.h.
// einen Knoten mit seinen Constraints

class no_cs {

  int       node;      // Nummer des Knotens im Array Nodes
  int       OA_cons;   // OA-Constraint des Knotens
  set<int>* SA_cons;   // SA-Constraint des Knotens

public:

  no_cs() : node(0), OA_cons(0), SA_cons(nil) {};
  no_cs(int n) : node(n), OA_cons(Nodes[n].OA), SA_cons(Nodes[n].SA) {};

  void remove_OA() { OA_cons = 0; }
  void change_SA(const no_cs& no) { SA_cons = no.SA_cons; }
 
  int OA() const { return OA_cons; }
  int adjoinable_at(const no_cs&) const;

  operator int() { return node; }
  friend ostream& operator<< (ostream&, const no_cs&);
  friend int compare(no_cs const &, no_cs const &);
};


// liefert 1 <=> akt. Knoten ist an Knoten "n1" adjungierbar; 0 sonst
int no_cs::adjoinable_at(const no_cs& n1) const
{
  // wenn es kein SA-Constraint gibt, dann gibt es sowieso keine Probleme
  if (!n1.SA_cons) return 1;
  // ansonsten mu der akt. Knoten in der SA-Menge von n1 sein
  return n1.SA_cons->member(node);
}

ostream& operator<< (ostream& out, const no_cs& nc)
{
  out << nc.node;
  if (nc.OA_cons) out << "#OA";
  if (nc.SA_cons) {
    out << "{ ";
    int cn;
    forall(cn,*nc.SA_cons) {
      out << cn << ' ';
    }
    out << "} ";
  }
  return out;
}

// LEDA braucht eine compare-Fkt., um festzustellen, ob das Element schon in
// der Menge enthalten ist.
// ( n1==n2: 0, n1<n2: -1, n1>n2: +1 )
// (Beim Vergleich der SA-Constraints findet nur ein "flaches" Compare statt,
//  d.h. es werden die Pointer und nicht die Mengen selbst verglichen; da sich
//  die Mengen nach der Init-Fkt. nicht mehr ndern, sollte das gengen.)
int compare(no_cs const &n1, no_cs const &n2)
{
  int comp;
  if ( comp = compare(n1.node, n2.node) ) return comp;
  if ( comp = compare(n1.OA_cons, n2.OA_cons) ) return comp;
  return compare(n1.SA_cons, n2.SA_cons);
}

// *** Die A[i,j,k,l]-Matrix selbst: ***

const int n_max = 10; // max. Lnge der Eingabe

set<no_cs> A[n_max+1][n_max+1][n_max+1][n_max+1];


// ***************************************************************************
// Initialisierungs- und Datei-Routinen

// Kommentare (; ...) im stream "in" berlesen
void eat_comment(ifstream& in)
{
  char ch;
  char dummy[256];

  do {
    in >> ch;
    if (ch == ';') {
      in.get(dummy,256,'\n'); 
      //cerr << "com: " << dummy << '\n';
    }  
    else
      in.putback(ch);
  } while ( ch==';' );

}

// Grammatik aus der Datei mit dem Pfad "fn" lesen
void read_gram(const char* fn)
{
  ifstream in(fn);
  if (!in) { cerr << "Fehler beim ffnen!\n"; exit(1); }

  eat_comment(in);
  in >> gram_name;           cerr << gram_name;
  eat_comment(in);
  in >> no_num;              cerr << " (" << no_num << " Knoten)\n";
  eat_comment(in);

  Nodes = new node[no_num+1];

  int i;
  for(i=1; i<=no_num; i++) {
    eat_comment(in);
    in >> Nodes[i];
  }
}

// ***************************************************************************
// Verschiedenes

void print_set(int i, int j, int k, int l)
{
  cout << '(' << i << ',' << j << ',' << k << ',' << l << "): ";
  no_cs cn;
  forall(cn,A[i][j][k][l]) { cout << cn << ' '; }
}

int max(int a, int b)
{
  return (a > b) ? a : b;
}
 
// ***************************************************************************
// Hier ist der (modifizierte) Vijay-Shankar Algorithmus

void case1(int i, int j, int k, int l)
{
  if (j == k) return; // !5

  int m,p;

  for(m=k; m<=l-1; m++)
    for(p=m; p<=l; p++) {

      no_cs n1,n2;
      forall(n1,A[i][j][k][m]) {
	if ( Nodes[n1].anc_of_foot )
	  forall(n2,A[m][p][p][l]) {
	    if ( Nodes[n1].right_sibl == n2 && !n1.OA() && !n2.OA() ) {
	      no_cs n(Nodes[n1].father);
	      A[i][j][k][l].insert(n);
	      cout << "C1[" << m << ',' << p << "]: " << n << ' ';
	    }
	  }	  
      }

    }
}

void case2(int i, int j, int k, int l)
{
  if (j == k) return; // !5

  int m,p;

  for(m=j; m>=i+1; m--) // !1
    for(p=m; p>=i; p--) {

      no_cs n1,n2;
      forall(n1,A[m][j][k][l]) {
	if ( Nodes[n1].anc_of_foot )
	  forall(n2,A[i][p][p][m]) {
	     if ( Nodes[n1].left_sibl == n2 && !n1.OA() && !n2.OA() )  {
	       no_cs n(Nodes[n1].father);
	       A[i][j][k][l].insert(n);
	       cout << "C2[" << m << ',' << p << "]: " << n << ' ';	
	     }
	  }  
      }

    }
}

void case3(int i, int j, int k, int l)
{
  if (j != k) return; // !4

  int m,p;

  for(m=max(j,i+1); m<=l-1; m++)
    for(p=m; p<=l; p++) {

	no_cs n1,n2;
	forall(n1,A[i][j][j][m]) {
	  forall(n2,A[m][p][p][l]) {
	    if ( Nodes[n1].right_sibl == n2 && !n1.OA() && !n2.OA() ) {
	      no_cs n(Nodes[n1].father);
	      A[i][j][j][l].insert(n);
	      cout << "C3[" << m << ',' << p << "]: " << n << ' ';	
	    }
	  }  
	}
    }

}

void case4(int i, int j, int k, int l)
{
  set<no_cs> new_no;
  no_cs n1;

  forall(n1,A[i][j][k][l]) {
    no_cs n = n1; // (sicherheitshalber)
    while ( !Nodes[n].left_sibl && !Nodes[n].right_sibl &&
	    Nodes[n].father && !n.OA() ) {
      n = Nodes[n].father;
      if (!A[i][j][k][l].member(n)) {
	new_no.insert(n);
	cout << "C4: " << n << ' ';
      }
    }
  }
  
  forall(n1,new_no) { 
    A[i][j][k][l].insert(n1);
  }  
}

void case5(int i, int j, int k, int l)
{
  set<no_cs> new_no;
  no_cs n1,n2;

  int m,p;

  for(m=i; m<=j; m++)
    for(p=k; p<=l; p++)
      if (i<m || p<l) { // !3

	forall(n2,A[i][m][p][l]) {
	  if ( Nodes[n2].type == root_aux ) {
	    char sym = Nodes[n2].symbol;
	    forall(n1,A[m][j][k][p]) {
	      if ( Nodes[n1].symbol == sym && n2.adjoinable_at(n1) ) {
		no_cs n1_ = n1; // (sicherheitshalber)
		n1_.remove_OA();
		n1_.change_SA(n2);
		new_no.insert(n1_);
		cout << "C5[" << m << ',' << p << "]: " << n1_ << ' ';	
	      }
	    }
	  }      
        }

      }

  forall(n1,new_no) {
    A[i][j][k][l].insert(n1);
  }
}

int vs(const char* inp)
{
  int i,j,k,l,n;
  
  n = strlen(inp);

  cout << '\n' << "Parsen der Eingabe '" << inp << "' (n=" << n << ") " \
       << "mit der Grammatik '" << gram_name << "' :\n\n";

  // step 1,2
  for(i=0; i<=n-1; i++) {
    int cn; for(cn=1; cn<=no_num; cn++)
      if (Nodes[cn].type == term && inp[i] == Nodes[cn].symbol) {
	A[i][i+1][i+1][i+1].insert(cn); 
      }
  }  

  // step 3-5
  for(i=0; i<=n; i++) // !2
    for(j=i+1; j<=n; j++) { // !6
      int cn; for(cn=1; cn<=no_num; cn++)
	if (Nodes[cn].type == foot) {
	  A[i][i][j][j].insert(cn);
	}
    }  

  // step 6-14
  for(l=0; l<=n; l++)
    for(i=l; i>=0; i--)
      for(j=i; j<=l; j++)
	for(k=l; k>=j; k--) {
          print_set(i,j,k,l);

	  case1(i,j,k,l);
	  case2(i,j,k,l);
	  case3(i,j,k,l);
	  case4(i,j,k,l); case5(i,j,k,l); // !3
	  case4(i,j,k,l);

	  cout << '\n';
	}  

  // step 15
  for(j=0; j<=n; j++) {
    no_cs cn;
    forall(cn,A[0][j][j][n]) {
      if ( Nodes[cn].type == root_ini && !cn.OA() ) {
	return 1;
      }
    }
  }
  return 0;
}

// ***************************************************************************

void main(int argc, char* argv[])
{
  char inp[50];
  char fn[50];

  cout << "VS Version " << VERSION << "\n";

  if (argc >= 2) 
    strcpy(fn, argv[1]);
  else {
    cerr << "Namen der TAG-Grammatik eingeben: "; cin >> fn;
  }

  read_gram(fn);

  if (argc >= 3) 
    strcpy(inp, argv[2]);
  else {
    cerr << "Zu parsenden String eingeben: "; cin >> inp;
  }
  if (strlen(inp) > n_max) {
    cerr << "Eingabe zu lang!\n"; exit(2);
  }

  if (vs(inp)) cout << "\nEingabe akzeptiert!\n";
  else         cout << "\nEingabe abgelehnt!\n";
}
